import json
import os
import sys
import time

import boto3
import click

PRL_PATH = os.path.expanduser("~/.prl")
CREDS_PATH = os.path.join(PRL_PATH, "creds.json")
VALS_ENV = os.getenv("VALS_ENV")

DEFAULT_REGION = "us-east-1"


def get_client_id(in_europe: bool, using_api_key: bool):
    if using_api_key:
        if in_europe:
            return "6cv1hchrihac7dtsmjac1dukn6"
        elif VALS_ENV in ["LOCAL", "DEV"]:
            return "7scu563gabte768gtml5v5uids"
        else:
            # Normal Prod user pool
            return "6t1s1a2g43ggqkn8timajdl0nn"
    else:
        if in_europe:
            return "4asi3qr1jga1l1kvc6cqpqdsad"
        elif VALS_ENV in ["LOCAL", "DEV"]:
            return "59blf1klr2lejsd3uanpk3b0r4"
        else:
            # Normal Prod user pool
            return "7r5tn1kic6i262mv86g6etn3oj"


def get_region():
    if "PRL_REGION" in os.environ:
        return os.environ["PRL_REGION"]

    if os.path.exists(CREDS_PATH):
        with open(CREDS_PATH, "r") as f:
            auth_dict = json.load(f)

        if "region" in auth_dict:
            return auth_dict["region"]

    return DEFAULT_REGION


def get_auth_token():
    if not os.path.exists(CREDS_PATH):
        auth_dict = {}
    else:
        with open(CREDS_PATH, "r") as f:
            auth_dict = json.load(f)

    if "PRL_REGION" in os.environ:
        region_name = os.environ["PRL_REGION"]
    elif "region" in auth_dict:
        region_name = auth_dict["region"]
    else:
        region_name = DEFAULT_REGION

    client = boto3.client("cognito-idp", region_name=region_name)

    if "access_expiry" not in auth_dict or time.time() > auth_dict["access_expiry"]:
        # API Key is specified in environment
        if "VALS_API_KEY" in os.environ:
            refresh_token = os.environ["VALS_API_KEY"]
            client_id = get_client_id(region_name == "eu-north-1", True)
        # We're using the prl login workflow.
        elif "refresh_token" in auth_dict:
            refresh_token = auth_dict["refresh_token"]
            client_id = auth_dict["client_id"]
        # No refresh token, so not logged in
        else:
            click.echo("Not authenticated. Run the command: prl login.")

        # If enough time has elapsed, we need to refresh the token
        try:
            response = client.initiate_auth(
                AuthFlow="REFRESH_TOKEN_AUTH",
                AuthParameters={"REFRESH_TOKEN": refresh_token},
                ClientId=client_id,
            )
        except Exception as e:
            click.echo(
                "Either your session has expired or an invalid API Key was provided. Run prl login, or update your VALS_API_KEY environment variable."
            )
            sys.exit()

        auth_dict = {
            **auth_dict,
            "access_token": response["AuthenticationResult"]["AccessToken"],
            "id_token": response["AuthenticationResult"]["IdToken"],
            "access_expiry": int(
                time.time() + response["AuthenticationResult"]["ExpiresIn"] - 10
            ),
        }
        auth_json = json.dumps(auth_dict, indent="\t")

        if not os.path.exists(PRL_PATH):
            os.makedirs(PRL_PATH, mode=0o770)

        # Store the new access token
        with open(CREDS_PATH, "w") as f:
            os.chmod(CREDS_PATH, mode=0o770)
            f.write(auth_json)

    return auth_dict["access_token"]



import json
import os
import time

import boto3
import click
from vals.sdk.auth import CREDS_PATH, PRL_PATH, get_client_id


@click.command(name="login")
def login_command():
    """
    Authenticate with Vals CLI
    """
    in_eu = click.confirm("Are you located in Europe?")
    region = "eu-north-1" if in_eu else "us-east-1"
    username = click.prompt("email")
    password = click.prompt("password", hide_input=True)

    client_id = get_client_id(in_eu, False)

    client = boto3.client("cognito-idp", region_name=region)

    # TODO: Error handling
    response = client.initiate_auth(
        AuthFlow="USER_PASSWORD_AUTH",
        AuthParameters={"USERNAME": username, "PASSWORD": password},
        ClientId=client_id,
    )

    auth_dict = {
        "refresh_token": response["AuthenticationResult"]["RefreshToken"],
        "access_token": response["AuthenticationResult"]["AccessToken"],
        "id_token": response["AuthenticationResult"]["IdToken"],
        "client_id": client_id,
        "region": region,
        "access_expiry": int(
            time.time() + response["AuthenticationResult"]["ExpiresIn"] - 10
        ),
    }
    auth_json = json.dumps(auth_dict, indent="\t")

    if not os.path.exists(PRL_PATH):
        os.makedirs(PRL_PATH)
        os.chmod(PRL_PATH, mode=0o770)

    with open(CREDS_PATH, "w") as f:
        os.chmod(CREDS_PATH, mode=0o770)
        f.write(auth_json)



from functools import wraps
from time import time
from typing import Callable

import attrs
import click
import requests
from openai import OpenAI
from tqdm import tqdm
from vals.cli.suite import pull_suite, update_suite
from vals.sdk.auth import get_auth_token
from vals.sdk.run import get_run_url, start_run
from vals.sdk.util import be_host

in_tokens = 0
out_tokens = 0


def parse_test_suite_id_from_url(test_suite_url: str) -> str:
    start_index = test_suite_url.find("test_suite_id=") + len("test_suite_id=")
    return test_suite_url[start_index:]


def read_file(file_id: str):
    response = requests.post(
        url=f"{be_host()}/read_file_text/?file_id={file_id}",
        headers={"Authorization": get_auth_token()},
    )
    return response.text


def run_evaluations(
    test_suite_url: str,
    generate_fn: Callable[[str], str],
    description="Ran automatically using the PRL SDK",
    maximum_threads=4,
    verbosity=1,
    model_under_test="sdk",
    **kwargs,
):
    test_suite_id = parse_test_suite_id_from_url(test_suite_url)

    suite_data = pull_suite(test_suite_id, include_id=True)

    if verbosity == 0:
        iterator = suite_data["tests"]
    else:
        iterator = tqdm(suite_data["tests"])

    global in_tokens, out_tokens
    metadata = {}

    uses_files = False
    for test in suite_data["tests"]:
        if "file_id" in test and test["file_id"] is not None:
            uses_files = True

    for test in iterator:
        start = time()

        if uses_files:
            if "file_id" in test and test["file_id"] is not None:
                file_id = test["file_id"]
                file_text = read_file(file_id)
            else:
                file_text = ""
            fixed_output = generate_fn(test["input_under_test"], file_text, **kwargs)
            test["fixed_output"] = fixed_output
        else:
            test["fixed_output"] = generate_fn(test["input_under_test"], **kwargs)

        end = time()
        metadata[test["id"]] = {
            "in_tokens": in_tokens,
            "out_tokens": out_tokens,
            "duration_seconds": end - start,
        }
        in_tokens = 0
        out_tokens = 0

    update_suite(test_suite_id, suite_data)
    run_id = start_run(
        test_suite_id,
        {
            "use_fixed_output": True,
            "description": description,
            "maximum_threads": maximum_threads,
            "model_under_test": model_under_test,
            **kwargs,
        },
        metadata_map=metadata,
    )
    run_url = get_run_url(run_id)

    if verbosity >= 1:
        click.secho(
            "Successfully updated test suite with new fixed outputs and started a new run.",
            fg="green",
        )
        click.secho(run_url, bold=True)

    return run_id


def wrap_chatcompletion(func: Callable):
    @wraps(func)
    def wrapper(**kwargs):
        response = func(**kwargs)
        global in_tokens, out_tokens

        in_tokens += response.usage.prompt_tokens
        out_tokens += response.usage.completion_tokens

        return response

    return wrapper


# External Facing
def patch(client: OpenAI):
    client.chat.completions.create = wrap_chatcompletion(client.chat.completions.create)
    return client
